//
// Copyright (c) 2012 Secret Geometry, Inc.. All rights reserved.
//

#ifndef CONTEXT_HPP
#define CONTEXT_HPP

#include "CellGridBuffer.hpp"
#include "Types.hpp"
#include "Utilities.hpp"

#define RESPONSE_BUFFER_SIZE 256 

namespace term {

// 
// Context
// 

class Context
{
  public:
    
    //
    Context(int width, int height, int scrollbackHeight, BeepCallback beepFunc, ResizeCallback resizeFunc);
    
    // Return to starting state.
    void reset();
    
    // Fill the grid with spaces.
    void clearBuffer();
    
    //
    void fillWithTestPattern();
    
    //
    void writeRawData(const Cell* buffer, int count);
    
    // Set the grid's dimensions.
    void setColumnCount(int cols);
    void setRowCount(int rows);
    void resize(int width, int height, bool shouldCallback = false);
    
    // Set the user-friendly title.
    void setTitle(const char* title);
    
    // Beep.
    void bell();
    
    // Write text at the current cursor position.
    void echo(uint8_t* s, int length);
    void echo(uint32_t c);
    
    // Add an accent to the previous character written.
    void addDiacritic(uint32_t diacritic);
    
    // 
    void setUseLineDrawingCharacters(bool b);
    
    // Grid accessor.
    CellGridBuffer& getGrid();
    
    // Cursor accessors.
    const CellCursor getCursor();
    void setCursor(CellCursor c);
    
    // Manage the cursor stack.
    void saveCursor();
    void restoreCursor();
    
    // Move the cursor if possible.
    void moveCursorUp(int n = 1);
    void moveCursorDown(int n = 1);
    void moveCursorLeft(int n = 1);
    void moveCursorRight(int n = 1);
    void moveCursorForward();
    void moveCursorBackward();
    void moveCursorToFarLeft();
    void moveCursorToNextTabStop();
    void moveCursorToRowCol(int row, int col);
    
    // Move the cursor and scroll when the margins are reached.
    void pushCursorUp();
    void pushCursorDown();
    void pushCursorLeft();
    void pushCursorRight();
    void pushCursorToNewLine();
    void pushCursorToFirstColOfNewLine();
    
    // Create a new cursor by shifting a copy of the given cursor.
    CellCursor cursorByMovingCursorForward(CellCursor c);
    CellCursor cursorByMovingCursorBackward(CellCursor c);
    
    // Sets the margins inside of which scrolling happens.
    // The parameters are 1-indexed and must conform to (1 <= top < bottom <= height).
    // Out of range values will be clamped.
    void setTopAndBottomMargins(int top, int bottom);
    
    void scrollUp(int n = 1);
    void scrollDown(int n = 1);
    
    // Set the color and style that will be used when writing characters.
    // The args array can contain multiple settings (eg. blue and underline).  See the specs for details.
    void selectGraphicRendition(const ControlArgs& args);
    
    // Erases characters in the current line using the given mode.
    void eraseInLine(int mode);
    
    // Erases characters in the grid using the given mode.
    void eraseInDisplay(int mode);
    
    // Erases the given number of characters without moving the cursor.
    void eraseCharacters(int count);
    
    // Deletes a single character behind the cursor.  All subsequent characters will be shifted backward. 
    void deleteBackwards();
    
    //
    void insertBlankCharacters(int count);
    
    // Deletes characters in the current line.
    void deleteCharacters(int count);
    
    // 
    void insertLines(int count);
    
    // 
    void deleteLines(int count);
    
    //
    void setTabStop();
    void clearTabStop();
    void clearAllTabStops();
    
    // Enable/disable the given mode.
    void enableMode(const ControlArgs& modes, bool questionMark);
    void disableMode(const ControlArgs& modes, bool questionMark);
    
    // Requests a terminal identification string.
    void requestDeviceAttributes();
    
    // 
    void requestDeviceStatusReport(int type);
    
    // Holds the response to any requests made during the last update.
    const char* getResponse();
    
  private:
    
    void scrollUp(int top, int bottom, int rowCount);
    void scrollDown(int top, int bottom, int rowCount);
    
    CellGridBuffer grid;
    CellCursor     cursor;
    CellCursor     backupCursor;
    Cell           prototypeCell;
    Cell           backupPrototypeCell;
    int            topMargin;
    int            bottomMargin;
    IntSet         tabStops;
    bool           insert;
    bool           autoWrap;
    bool           autoReverseWrap;
    bool           originMode;
    bool           wideColumnMode;
    bool           lineFeedCarriageReturn;
    bool           useLineDrawingCharacters;
    bool           needsWrap;
    BeepCallback   beepCallback;
    ResizeCallback resizeCallback;
    Request        request;
    char           response[RESPONSE_BUFFER_SIZE];
};

} // namespace term

#endif // CONTEXT_HPP
